/*
 * Copyright (C) 2008 The Trustees of Columbia University in the City of New York. 
 * All rights reserved.
 *
 * See the file LICENSE included in this distribution for details.
 */

#include "serverbw.h"
#include "gettimeofday.h"
#include "trace.h"
#include "socketi.h"

#ifdef _LINUX_
#include <arpa/inet.h>
#endif

#ifdef WIN32
#include <winsock.h>
#endif

#include <string>
#include <map>
using namespace std;



ServerBW::ServerBW() {
  m_iTermFlag=1;  //1 means continue
  m_isRunning=0;  //is the measurement running or waiting for connection

  gettimeofday(&m_endtime, NULL);

  allocmutex_i(&m_mutex_running);
  
}//ServerBW


ServerBW::~ServerBW() {
  deallocmutex_i(&m_mutex_running);
}//~ServerBW


void ServerBW::Initialize(uint16 tcpserverport, int32 mode, int32 pktlen, int32 clientagnost, int32 both, 
			  int32 (*callback)(uint32 sbw, uint32 rbw, int32 result, 
					    int32 retryafter, int32 isClient, void* pApp), 
			  void* pApp) {
  m_usTcpPort = tcpserverport;
  m_mode = mode;
  m_iPktLen = pktlen;
  m_iClientAgnost = clientagnost; //whether measurement test can be from any client
  m_iBoth = both;                 //perform bw measurements both ways

  m_callback = callback;
  m_pApp = pApp;

  m_isRunning = 0;  //By default, we are not running
}//ServerBW::Initialize


/*
  -2 shutting down
  -1 error
  0 cannot accept connection. a measurement in progress
  >1 the fd of the socket
 */
int32 ServerBW::HandleNewConnection() {
  int connection;
  char8 pchIP[IPADDRLEN*4];
  uint16 cport;

  char temp[10];
  string str;
  struct sockaddr_in staddr;  
  int optarg;
  int32 len;
  struct bwconn stbwconn;

  
  //connection
  connection = accept(m_iTcpSock, NULL, NULL);

  //If cannot accept
  if(connection < 0) {
    TRACE(-1, "accept\n", "");
    return -1;
  }

  //Check the IP address. If same as ours, we are trying to close the connection
  //the IP address can be something other than 127.0.0.1
  len=sizeof(staddr);
  //get the IP address and port number of the incoming connection
  getpeername(connection, (struct sockaddr*) &staddr, (socklen_t*)&len);  
  strncpy(pchIP, inet_ntoa(staddr.sin_addr), IPADDRLEN);
  cport=htons(staddr.sin_port);

  TRACE(3, "ServerBW::HandleNewConnection...IP: %s port: %d expected IP: %s port: %d\n", pchIP, cport, m_pchClientIP, m_usClientPort);

  //printf("IP address closing: %s\n", arr);
  if(0==strncmp("127.0.0.1", pchIP, IPADDRLEN) ) {
    close(connection);
    m_iTermFlag=0;
    return -2;
  }

  if(1==m_SL.size() ) {
    TRACE(2, "ServerBW::HandleNewConnection...a measurement in progress\n", "");
    close(connection);
    return 0;
  }

  //We need to check for a specific client
  if(0==m_iClientAgnost) {
    //If this client was not negotiated, immediately close the connection.
    if(0 != strncmp(pchIP, m_pchClientIP, IPADDRLEN) && cport != m_usClientPort)  {
      TRACE(2, "ServerBW::HandleNewConnection...connection from non-negotiated client\n", "");
      close(connection);
      return -1;
    }
  }//end if


  ////////////////////////
  //Set socket options
  ////////////////////////

  //(1) set nonblocking
  setsockopti_tcp(connection, S_TCP_NONBLOCK, NULL, 0);

  /*
  //(2) snd buffer  
  optarg=SOCKBUFLEN;
  len=sizeof(optarg); 
  setsockopti_tcp(m_iTcpSock, S_SNDBUF, (void*)&optarg, len);

  //(3) rcv buffer
  optarg=SOCKBUFLEN*10;
  len=sizeof(optarg); 
  setsockopti_tcp(m_iTcpSock, S_RCVBUF, (void*)&optarg, len);
  */

  //(4) No Delay
  optarg=1;
  len=sizeof(optarg);
  setsockopti_tcp(connection, S_TCP_NODELAY, (void*)&optarg, len);

  //(5) QUICK ACKs
  optarg=1;
  len=sizeof(optarg);
  setsockopti_tcp(connection, S_TCP_QUICKACK, (void*)&optarg, len);

  ///////////////////////////
  //Set socket options ends
  ///////////////////////////


  //////////////////////////
  //get socket options to test
  /////////////////////////
  /*optarg=0;
  len=sizeof(optarg);
  getsockopti_tcp(connection, S_RCVBUF, (void*)&optarg, &len);
  //printf("rcvbuf: %d ", optarg);

  optarg=0;
  len=sizeof(optarg);
  getsockopti_tcp(connection, S_TCP_QUICKACK, (void*)&optarg, &len);
  //printf("quick ack :%d\n", optarg);
  */

  len=sizeof(staddr);
  getpeername(connection, (struct sockaddr*) &staddr, (socklen_t*)&len);

  strncpy(pchIP, inet_ntoa(staddr.sin_addr), 30);
  sprintf(temp, "%d", htons(staddr.sin_port));
  str = pchIP;
  str += ":";
  str += temp;
  
  TRACE(3, "ServerBW::HandleNewConnection...%s\n", (char*)str.c_str());

  stbwconn.sockfd = connection;
  gettimeofday(&stbwconn.starttime, NULL);
  stbwconn.stoptime=addtimeval(stbwconn.starttime, BWTESTLEN, 0);
  stbwconn.state=BW_MEASURE;
  stbwconn.bytesrcvd=0;
  stbwconn.bytessent=0;
    
  m_SL[str] = stbwconn;

  //printf("---------------------------------------\n");
  //printf("New connection: %s\n", (char*)str.c_str() );

  TRACE(3, "ServerBW::HandleNewConnection...size: %d\n", m_SL.size() );

  return connection;
}//ServerBW::HandleNewConnection



int32 ServerBW::ReadSocks() {
  
  map<string, bwconn>::iterator itr, itr1;
  struct timeval curtime, difftime;
  int32 ret;
  //int32 optarg, len;
 
  //TRACE(2, "ServerBW::ReadSocks...map size: %d\n", m_SL.size() );
 
  for(itr=m_SL.begin(); itr!=m_SL.end(); itr++) {

    //TRACE(2, "ServerBW::ReadSocks...calling recv for TCP size: %d\n", m_SL.size() );
    if(FD_ISSET(itr->second.sockfd, &m_readfds)) {

      //current time
      gettimeofday(&curtime, NULL);
      
      difftime=difftimeval(itr->second.stoptime, curtime);
      if(0==difftime.tv_sec && 0==difftime.tv_usec) {

	//If measuring, then terminate
	if(itr->second.state == BW_MEASURE) {
	  //printf("Measurement Completed...\n");
	  //test completed. get the end time
	  gettimeofday(&itr->second.closetime, NULL);
	  close(itr->second.sockfd);
	  TRACE(2, "ServerBW::ReadSocks...closing connection .. timeup\n", "");
	  m_rSL[itr->first] = itr->second;
	  break;
	}
      }//end if
      

      //TRACE(2, "ServerBW::ReadSocks...recv\n", "" );
      ret=recv(itr->second.sockfd, (char*)(itr->second.buf), m_iPktLen, 0);
      //TRACE(2, "ServerBW::ReadSocks...ret: %d\n", ret );

      if(ret > 0) {
	itr->second.bytesrcvd += ret;

      } else if(0==ret || -1==ret) {
	gettimeofday(&itr->second.closetime, NULL);
	close(itr->second.sockfd);
	TRACE(2, "ServerBW::ReadSocks...closing connection\n", "");
	m_rSL[itr->first] = itr->second;
      }

    }//end if

  }//end for

    
  //now remove the closed connections from the map
  for(itr=m_rSL.begin(); itr!=m_rSL.end(); itr++) {
    itr1 = m_SL.find(itr->first);    

    if(itr1 != m_SL.end() ) {      

      //Critical section begins
      lockmutex_i(&m_mutex_running);      

      ReportResults(itr1->second, 1);

      //printf("destroyed tcp connection\n");
      m_SL.erase(itr1);

      //the measurement is closed
      m_isRunning=0;

      //Critical section ends
      unlockmutex_i(&m_mutex_running);
    }
  }//end for

  //clear the remove list
  m_rSL.clear();


  TRACE(2, "ServerBW::ReadSocks...Ends\n", "");
  
  return 1;
}//ReadSocks


int32 ServerBW::WriteSocks() {
  map<string, bwconn>::iterator itr, itr1;
  int32 ret=0;

  //write data
  TRACE(2, "ServerBW::WriteSocks...size: %d\n", m_SL.size() );

  for(itr=m_SL.begin(); itr!=m_SL.end(); itr++) {
    TRACE(2, "ServerBW::WriteSocks...calling send() for TCP\n", "");
    
    if(FD_ISSET(itr->second.sockfd, &m_writefds)) {
      ret=send(itr->second.sockfd, itr->second.buf, m_iPktLen, 0);
      if(ret > 0) {
	itr->second.bytessent += ret;
      } else {
	//if we cannot write, close immediately
	gettimeofday(&itr->second.closetime, NULL);
	close(itr->second.sockfd);
	TRACE(2, "ServerBW::WriteSocks...closing connection\n", "");
	m_rSL[itr->first] = itr->second;

      }
    }//end if
  }//end for


  //now remove the closed connections from the map
  for(itr=m_rSL.begin(); itr!=m_rSL.end(); itr++) {
    itr1 = m_SL.find(itr->first);    

    if(itr1 != m_SL.end() ) {      

      //Critical section begins
      lockmutex_i(&m_mutex_running);      

      ReportResults(itr1->second, 1);

      //printf("destroyed tcp connection\n");
      m_SL.erase(itr1);

      //the measurement is closed
      m_isRunning=0;

      //Critical section ends
      unlockmutex_i(&m_mutex_running);
    }
  }//end for

  
  return 1;
}//WriteSocks


int32 ServerBW::Run() {
  int32 ret;
  ret=RunTCP();
  return ret;
}//Run


/*
  1 if running
  0 if not running
  retyafter: if running, retry after this much time
 */
int32 ServerBW::isRunning(uint32& retryafter) {
  int32 size, ret;
  struct timeval curtime, difftime;

  //lockmutex_i(&m_mutex_running);
  ret=trylockmutex_i(&m_mutex_running);
  if(0==ret) {
    return 0;
  } else if(-1==ret) {
    return -1;
  }

  //We acquired the lock

  size=m_SL.size();

  //If we are running
  if(size>0) {
    retryafter=BWTESTLEN;
    unlockmutex_i(&m_mutex_running);
    return 1;
  }

  //if we are not running
  if(0==m_isRunning) {
    unlockmutex_i(&m_mutex_running);
    return 0;
  }

  //If list size is 0 and if we m_isRunning is set, check if we 
  //have not received a connection
  if(1==m_isRunning && 0==size) {
    gettimeofday(&curtime, NULL);

    //where is m_endtime set?
    difftime=difftimeval(m_endtime, curtime);
    if(0==difftime.tv_sec && 0==difftime.tv_usec) {      
      retryafter=0;
      unlockmutex_i(&m_mutex_running);
      return 0;
    }
  }//end if

  unlockmutex_i(&m_mutex_running);

  retryafter=BWTESTLEN;
  return 1;
}//ServerBW::isRunning


uint16 ServerBW::getServerPort() {
  return m_usTcpPort;
}//getServerPort


int32 ServerBW::RunTCP() {
  
  int32 sockready=0;
  int32 ret;
  struct sockaddr_in staddr;
  int32 len;
  //struct timeval timeout, curtime, difftime;
  
  //Create TCP socket
  m_iTcpSock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  
  //bind it to the port
  if(-1 == bindi(m_iTcpSock, m_usTcpPort) ) {
    TRACE(-1, "ServerBW::RunTCP...Error in TCP bind\n", "");
    m_usTcpPort=0;
    bindi(m_iTcpSock, m_usTcpPort); //bind to any arbitrary available port    

    len=sizeof(staddr);
    //get the IP address and port number where the bandwidth port is bound
    getsockname(m_iTcpSock, (struct sockaddr*) &staddr, (socklen_t*)&len);
    m_usTcpPort=htons(staddr.sin_port);

    TRACE(2, "ServerBW::RunTCP...port is: %u\n", m_usTcpPort);
  }//end if
  

  //TCP port listen
  ret = listen(m_iTcpSock, 2);

  while(m_iTermFlag) {

    BuildSelectList();

    sockready = select(m_sockmax, &m_readfds, &m_writefds, (fd_set*)0, NULL);
    
    if(FD_ISSET(m_iTcpSock, &m_readfds)) {
      TRACE(5, "ServerBW::RunTCP...a new TCP connection\n", "");
      HandleNewConnection();
    }


    //ret=recv(fd, (char*)(buf), m_iPktLen, 0);
    //TRACE(3, "ServerBW::RunTCP...size: %d ret: %d\n", m_SL.size(), ret );

    //printf("read socks\n");
    ReadSocks();
    
    if(1==m_iBoth) {
      WriteSocks();
    }
    
  }//end while


  //TODO: cleanup here

  return 1;
}//RunTCP


/*
  1 success
  -1 error
  0 already a measurement in progress
 */
int32 ServerBW::NewClient(char8* pchIP, uint16 port, int32 both) {
  int32 ret;
  struct timeval difftime, curtime;
  int32 isNewClient=0;

  //If there must be a specific client
  if(1==m_iClientAgnost) {
    TRACE(5, "ServerBW::NewClient...a specific client for agnostic listener\n", "");
    return 1;
  }

  ret=trylockmutex_i(&m_mutex_running);
  if(0==ret) {
    return 0;
  } else if(-1==ret) {
    return -1;
  }

  
  //We are able to obtain a lock
  //a measurement in progress? m_SL size should be greater than one
  if(m_SL.size() > 0) {
    TRACE(5, "ServerBW::NewClient...a measurement in progress\n", "");
    unlockmutex_i(&m_mutex_running);
    return 0;
  }
  

  //We had scheduled a measurement but we still do not have a connection
  if(1==m_isRunning && 0==m_SL.size() ) {
    gettimeofday(&curtime, NULL);
    difftime=difftimeval(m_endtime, curtime);    
   
    //If we have not received a connection, we can use a new client
    if(0==difftime.tv_sec && 0==difftime.tv_usec) {
      isNewClient=1;      
    }
  } else if (0==m_isRunning && 0==m_SL.size() ) {
    isNewClient=1;
  }
  

  if(1==isNewClient) {
    m_isRunning=1;   //a measurement is scheduled
    gettimeofday(&curtime, NULL);
    m_endtime=addtimeval(curtime, BWTESTLEN, 0); //timer. If we do not receive a connection, we can do another measurement.
    strncpy(m_pchClientIP, pchIP, IPADDRLEN);
    m_usClientPort=port;
    m_iBoth = both;  //measurements are both ways or one way

    TRACE(5, "ServerBW::NewClient...client IP: %s port: %d\n", m_pchClientIP, m_usClientPort);

    unlockmutex_i(&m_mutex_running);
    return 1;
  }

  unlockmutex_i(&m_mutex_running);
  return 0;
}//NewClient


/*
  -1 if error
  1 if successful
 */
int32 ServerBW::Terminate() {
  //establish a connection with itself
  //this will cause the select() to return
  int32 sockfd;
  struct sockaddr_in dest_addr, myaddr;
  
  //Create TCP socket
  sockfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  
  //bind UDP listener to UDP port
  myaddr.sin_family = AF_INET;
  myaddr.sin_addr.s_addr = inet_addr("127.0.0.1");
  myaddr.sin_port = htons(0);
  memset(&(myaddr.sin_zero), '\0', 8);

  if(-1==bind(sockfd, (struct sockaddr *)&myaddr, sizeof(myaddr)) ) {
    TRACE(2, "ServerBW::Terminate...Error in TCP bind\n", "");
    return -1;
  }

  //connect to the server to indicate that we are shutting down
  dest_addr.sin_family = AF_INET;            // host byte order
  dest_addr.sin_port = htons(m_usTcpPort);   // short, network byte order
  dest_addr.sin_addr.s_addr = inet_addr("127.0.0.1");
  memset(&(dest_addr.sin_zero), '\0', 8);    // zero the rest of the struct
  
  if(0 != ::connect(sockfd, (struct sockaddr *)&dest_addr, sizeof(struct sockaddr)) ) {
    TRACE(-1, "ServerBW::Terminate...Error in TCP connect()\n", "");
    return -1;
  }//end if

  close(sockfd);

  return 1;
  
}//Terminate


/*
  builds the read and write descriptor list
 */
void ServerBW::BuildSelectList() {
  map<string, bwconn>::iterator itr, itr1;

  FD_ZERO(&m_readfds);
  FD_ZERO(&m_writefds);

  FD_SET(m_iTcpSock, &m_readfds);  //read descriptors  
  
  m_sockmax = m_iTcpSock+1;
  
  for(itr=m_SL.begin(); itr!=m_SL.end(); itr++) {
    m_sockmax=itr->second.sockfd+1;
    FD_SET(itr->second.sockfd, &m_readfds);

    if(m_iBoth) {
      FD_SET(itr->second.sockfd, &m_writefds);
    }
  }//end for
    
}//BuildSelectList


int32 ServerBW::ReportResults(struct bwconn& stbwconn, int32 result) {

  struct timeval difftime;
  uint32 ms;
  uint32 sbw, rbw;
  
  difftime=difftimeval(stbwconn.closetime, stbwconn.starttime);

  ms=difftime.tv_sec*1000+difftime.tv_usec/1000;
  if(ms/1000 == 0) {
    sbw=0;
    rbw=0;
  } else {
    sbw=stbwconn.bytessent/ms*8*1000;
    rbw=stbwconn.bytesrcvd/ms*8*1000;
  }
  //printf("sent: %u rcvd: %u time: %d sent/s: %u rcvd/s: %u\n", stbwconn.bytessent, stbwconn.bytesrcvd, ms, sbw, rbw);  
  //printf("---------------------------------------\n\n");

  //second 0 is to indicate that we are server.
  m_callback(sbw, rbw, result, 0, 0, m_pApp);

  return 1;
}//ReportResults
