/*
 * Copyright (C) 2008 The Trustees of Columbia University in the City of New York. 
 * All rights reserved.
 *
 * See the file LICENSE included in this distribution for details.
 */

#ifndef _SERVERBW_H
#define _SERVERBW_H

#include "types.h"
#include "bwstruct.h"
#include "threadi.h"

#include <map>
using namespace std;

//#define IPADDRLEN 30

class ServerBW {
 private:
  uint16 m_usTcpPort;     //Tcp listening port
  int32 m_mode;           //0 for TCP 1 for UDP
  
  int32 m_iClientAgnost;  //0 if any client can be accepted. 
                          //1 for specific client negotiated using another protocol.  
  char8 m_pchClientIP[30];//Client IP
  uint16 m_usClientPort;  //Client port
  int32 m_iBoth;          //global. measurement is both way

  struct timeval m_endtime; 
  struct mutex_i m_mutex_running;
  int32 m_isRunning;      //a flag that a measurement is in progress or waiting for connection.

  int32 m_iTcpSock;       //Tcp listening socket
  int32 m_iUdpSock;       //TODO: Udp socket
  int32 m_iPktLen;        //Packet length. Should be an integral of MSS
  
  fd_set m_readfds;
  fd_set m_writefds;
  int32 m_sockmax;  

  int32 m_iTermFlag;      //terminate the listen loop. graceful way to shut down server bandwidth measurement thread

  map<string, bwconn> m_rSL;
  map<string, bwconn> m_SL;

  int32 (*m_callback)(uint32 sbw, uint32 rbw, int32 result, int32 retryafter, int32 isClient, void* pApp);
  void* m_pApp;


 private:
  int32 ReadSocks();
  int32 WriteSocks();
  int32 HandleNewConnection();
  void BuildSelectList();
  int32 RunTCP();

 public:
  ServerBW();
  ~ServerBW();
  void Initialize(uint16 tcpserverport, int32 mode, int32 pktlen, int32 clientagnost, int32 both, int32 (*callback)(uint32 sbw, uint32 rbw, int32 result, int32 retryafter, int32 isClient, void* pApp), void* pApp);
  int32 Run();  
  int32 isRunning(uint32& retryafter);
  int32 NewClient(char8* pchIP, uint16 port, int32 both);
  int32 Terminate();
  int32 ReportResults(struct bwconn& stbwval, int32 result);
  uint16 getServerPort();
};

#endif
