/*
 * Copyright (C) 2008-2009 The Trustees of Columbia University in the City of New York. 
 * All rights reserved.
 *
 * See the file LICENSE included in this distribution for details.
 */

#include "socketi.h"
#include "transport.h"

#include <string.h>

#ifdef _LINUX_
#include <sys/socket.h> //socket
#include <netinet/in.h> //hton, ntoh
#include <netinet/tcp.h> //tcpinfo
#include <sys/types.h>  //socket
#include <arpa/inet.h>  //inet_addr functions

#include <linux/sockios.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include <fcntl.h>
#endif

#ifdef WIN32
#include <winsock.h>
#endif


int32 bindi(int32 sockid, uint16 port) {
  struct sockaddr_in myaddr;
  int ret;

  //bind UDP listener to UDP port
  myaddr.sin_family = AF_INET;
  myaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  myaddr.sin_port = htons(port);
  memset(&(myaddr.sin_zero), '\0', 8);

  return bind(sockid, (struct sockaddr *)&myaddr, sizeof(myaddr));

  //port = ntohs(myaddr.sin_port);
  return ret;
}//bindi


int32 sendtoi(int32 sockfd, char* data, int32 size, char* ip, int32 port) {

  struct sockaddr_in dest_addr;
  int32 nbytes;

  dest_addr.sin_family = AF_INET;          // host byte order
  dest_addr.sin_port = htons(port);   // short, network byte order
  dest_addr.sin_addr.s_addr = inet_addr(ip);
  memset(&(dest_addr.sin_zero), '\0', 8);  // zero the rest of the struct

  nbytes = ::sendto(sockfd, data, size, 0, (struct sockaddr*)&dest_addr, sizeof(dest_addr));

  return nbytes;
}//sento


/*
 *sockfd: socket, timeoutval is value in ms
 */
int32 selectforsocket(int32 sockfd, int32 timeoutval) {
  fd_set selectfds;
  int32 sockmax;
  int readsocks=0;
  struct timeval timeout;
  
  //timeout value. set the timeout value in seconds
  timeout.tv_sec=timeoutval/1000;
  timeout.tv_usec=(timeoutval%1000) * 1000;

  FD_ZERO(&selectfds);
  FD_SET(sockfd, &selectfds);
  
  sockmax=sockfd+1;

  readsocks = select(sockmax, &selectfds, (fd_set*)0, (fd_set*)0, &timeout);

#ifdef WIN32
  if(0 == readsocks) {
    timeout.tv_sec = timeout.tv_usec = 0;
  }
#endif 

  //we do not have any data to read
  if(0 == timeout.tv_sec && 0 == timeout.tv_usec) {
    return 0;
  }//end if

  //we have data to read
  return 1;
}//selectforsocket




#ifdef WIN32
int close(int s) {
  return closesocket(s);
}//close

unsigned long inet_aton(const char* cp, struct in_addr* pin) {
  unsigned long temp;

  temp = inet_addr(cp);
  pin->s_addr = temp;

  return temp;
}
#endif

int32 getSockAddr_in(char *pchIP, int port, struct sockaddr_in *pstaddr) {
    pstaddr->sin_family=AF_INET;
    pstaddr->sin_addr.s_addr = inet_addr(pchIP);
    pstaddr->sin_port=htons(port);
    memset(pstaddr->sin_zero,'\0', sizeof(pstaddr->sin_zero));
	return 1;
}



int32 setsockopti_tcp(int32 sockid, enum SOCK_OPTIONS options, const void* optval, int optlen) {
  int32 ret=0;

  switch(options) {

  case S_TCP_NODELAY:
#ifdef _LINUX_
    ret = ::setsockopt(sockid, IPPROTO_TCP, TCP_NODELAY, optval, optlen);
#endif

#ifdef WIN32
    ret = ::setsockopt(sockid, IPPROTO_TCP, TCP_NODELAY, (const char*)optval, optlen);
#endif
    break;

  case S_TCP_QUICKACK:
#ifdef _LINUX_
    ret = ::setsockopt(sockid, IPPROTO_TCP, TCP_QUICKACK, optval, optlen);
#endif
    break;

  case S_SNDBUF:
#ifdef _LINUX_
    ret=::setsockopt(sockid, SOL_SOCKET, SO_SNDBUF, optval, optlen);
#else
    ret=::setsockopt(sockid, SOL_SOCKET, SO_SNDBUF, (const char*)optval, optlen);
#endif
	break;

  case S_TCP_INFO:
    break;

  case S_WQ:
    break;

  case S_RQ:
    break;

  case S_RCVBUF:
#ifdef _LINUX_
    ret=::setsockopt(sockid, SOL_SOCKET, SO_RCVBUF, optval, optlen);
#else
    ret=::setsockopt(sockid, SOL_SOCKET, SO_RCVBUF, (const char*)optval, optlen);
#endif
    break;

  case S_TCP_BLOCK:
#ifdef _LINUX_
    //ret=fcntl(sockid,F_SETFL,O_BLOCK);
#endif
#ifdef WIN32
    int32 NonBlock;
    NonBlock=0;
    ret=ioctlsocket(sockid, FIONBIO, (u_long*)&NonBlock);
#endif
    break;

  case S_TCP_NONBLOCK:
#ifdef _LINUX_
    ret=fcntl(sockid,F_SETFL,O_NONBLOCK);
#endif
#ifdef WIN32
    NonBlock=1;
    ret=ioctlsocket(sockid, FIONBIO, (u_long*)&NonBlock);
#endif
    break;

  default:
    break;
  }
  return ret;

}//setsockopti_tcp


int32 getsockopti_tcp(int32 sockid, enum SOCK_OPTIONS options, void* optval, int* optlen) {

  int ret=0;
  struct _tcp_info* ptcp;

#ifdef _LINUX_
  struct tcp_info stcp;
  socklen_t len = sizeof(stcp);
#endif

  switch(options) {
    
  case S_TCP_NODELAY:
#ifdef _LINUX_
    ret = ::getsockopt(sockid, IPPROTO_TCP, TCP_NODELAY, optval, (socklen_t*)optlen);
#endif

#ifdef WIN32
    ret = ::getsockopt(sockid, IPPROTO_TCP, TCP_NODELAY, (char*)optval, (socklen_t*)optlen);
#else
    *(int*)(optval)=1;
#endif
    break;

  case S_TCP_QUICKACK:
#ifdef _LINUX_
    ret = ::getsockopt(sockid, IPPROTO_TCP, TCP_QUICKACK, optval, (socklen_t*)optlen);
#else
    *(int*)(optval)=1;
#endif
    break;

  case S_TCP_INFO:
#ifdef _LINUX_
    ret = ::getsockopt(sockid, IPPROTO_TCP, TCP_INFO, &stcp, &len);
    if(0 == ret) {
      ptcp = (struct _tcp_info*)optval;
      ptcp->tcpi_snd_cwnd = stcp.tcpi_snd_cwnd;
      ptcp->tcpi_snd_ssthresh = stcp.tcpi_snd_ssthresh;
      ptcp->tcpi_ca_state = stcp.tcpi_ca_state;
      ptcp->tcpi_snd_mss = stcp.tcpi_snd_mss;
      ptcp->tcpi_rcv_mss = stcp.tcpi_rcv_mss;
      ptcp->tcpi_lost = stcp.tcpi_lost;
      ptcp->tcpi_total_retrans = stcp.tcpi_retrans;
      ptcp->tcpi_rtt = stcp.tcpi_rtt;
      //ptcp->tcpi_rcv_rtt = stcp.tcpi_rcv_rtt;
    }
#endif
    break;


  case S_RCVBUF:
#ifdef _LINUX_
    ret=::getsockopt(sockid, SOL_SOCKET,SO_RCVBUF, optval, (socklen_t*)optlen);
#else
    ret=::getsockopt(sockid, SOL_SOCKET,SO_RCVBUF, (char*)optval, (socklen_t*)optlen);
#endif
    break;

  case S_WQ:
    //get the write buffer queue size
#ifdef _LINUX_
    ioctl(sockid, SIOCOUTQ, optval);
#else
    //ioctlsocket(sockid, FIONWRITE, (u_long*)optval);
    *(int*)(optval)=1;
#endif
    break;

  case S_RQ:

#ifdef _LINUX_
    //get the read buffer queue size
    ioctl(sockid, SIOCINQ, optval);
#else
    ioctlsocket(sockid, FIONREAD, (u_long*)optval);
#endif
    break;

  default:
    break;

  }
  return ret;
}//getsockopti_tcp
