/*
 * Copyright (C) 2008-2009 The Trustees of Columbia University in the City of New York. 
 * All rights reserved.
 *
 * See the file LICENSE included in this distribution for details.
 */

#ifndef _TRANSPORT_H
#define _TRANSPORT_H

#ifndef _WENGO
#include "types.h"
#include "node.h"
#endif

#ifdef _WENGO
#include "p2pp/common/types.h"
#include "p2pp/node/node.h"
#include "p2pp/node/client.h"
#endif

#include <map>
using namespace std;

#ifdef _LINUX_
#include <sys/socket.h>
#endif

#ifdef WIN32
#include <winsock.h>
#endif

#define TIMEOUT 500000  //microseconds

class Node;
struct port_trans;


struct conninfo {
  int32 sockfd;
  uchar buf[65000];
  int32 msglen;
  int32 rcvdlen;

  conninfo() {
    msglen=0;
    rcvdlen=0;
  }
};

class Transport {

private:
  Node* m_pobjNode;

  int32 m_uiTcpPort;
  int32 m_uiUdpPort;
  int32 m_iTcpSock;
  int32 m_iUdpSock;

  fd_set sockfds;
  int32 sockmax;

  //socket maps
  map<string, conninfo> m_SL;      //listener socket
  map<string, int32> m_nSL;
  map<string, int32> m_rSL;
  
  //buffer to receive data in
  uchar m_buf[65000];

  uint32 m_curS;
  uint32 m_prevS;

  uint32 m_curR;
  uint32 m_prevR;
  struct timeval m_timeSR;

  int32 isConnected;
  int32 m_Listen;

private:
  void build_select_list();
  void handle_new_connection();
  void read_socks();
  int32 sockcon(char* ip, int32 port);

public:

  Transport();
  ~Transport();
  void setNode(Node* pobjNode);
  void Initialize(int32 iTcpPort, int32 iUdpPort); //Initialize with UDP and TCP listening port

  int32 Send(uchar* buf, uint32 len, char* ip, int32 port, transtype eType); //Send data to ip port. open a new connection if it does not exist.
  
  int32 getAddress(int32& num, char8 addr[][IPADDRLEN], uint16 port[], uchar ht[], uchar tt[]);

  int32 getAddrPort(map<string, port_trans>& addrmap);

  void getSendRecvBW(uint32& sbw, uint32& rbw);

  void Disconnect();

  static int32 gethostbyname(char* name, string& strIP);
  static int32 gethostbyaddr(char* ip, string& strHostName);  


  void ListenLoop() ;

  void SuspendListen(){
    m_Listen=0;
  }

  int getUDPSocket() {
    return m_iUdpSock;
  }
};

#endif
