/*
 * Copyright (C) 2008 The Trustees of Columbia University in the City of New York. 
 * All rights reserved.
 *
 * See the file LICENSE included in this distribution for details.
 */

#ifndef _SYSUTILS_H
#define _SYSUTILS_H

#include <string.h>

#ifdef _LINUX_
#include <unistd.h> //getopt
#endif


#ifdef WIN32
#include <windows.h>

int strcasecmp(const char* s1, const char* s2);
char getopt(int argc, char** argv, char* options);
void sleep(unsigned int seconds);
#endif

char *strtok_i(char *strToken, const char *strDelimit, char **context );


//Sources:
//[Linux] http://www.manpagez.com/man/2/send/
//[Windows XP] http://msdn.microsoft.com/en-us/library/ms740149(VS.85).aspx
//Default description is from Linux

/*
EACCES      WSAEACCES
EAGAIN                     The socket is marked non-blocking and the requested operation would block.
EBADF                      An invalid descriptor is specified.
ECONNRESET  WSAECONNRESET  A connection is forcibly closed by a peer.
EFAULT      WSAEFAULT      An invalid user space address is specified for a parameter.
EHOSTUNREACH WSAEHOSTUNREACH     The destination address specifies an unreachable host.
EINTR       WSAEINTR       [Linux] A signal interrupts the system call before any data is transmitted.
                           [Windows XP] A blocking Windows Sockets 1.1 call was canceled through WSACancelBlockingCall.
EMSGSIZE    WSAEMSGSIZE    The socket requires that message be sent atomically, and the size of the message to be sent makes this
                           impossible.
ENETDOWN    WSAENETDOWN    The local network interface used to reach the destination is down.
ENETUNREACH WSAENETUNREACH No route to the network is present.
ENOBUFS     WSAENOBUFS     The system is unable to allocate an internal buffer. The operation may succeed when buffers become available.
ENOTSOCK    WSAENOTSOCK    The argument socket is not a socket.
EOPNOTSUPP  WSAEOPNOTSUPP  socket does not support (some of) the option(s) specified in flags.
EPIPE                      The socket is shut down for writing or the socket is connection-mode and is no longer connected.  In the latter case, and if the socket is of type SOCK_STREAM, the SIGPIPE signal is generated to the calling thread.

*/

enum ERRORNOI {EI_NOTINITIALIZED=0, EI_ACCES, EI_AGAIN, EI_BADF, EI_CONNRESET, EI_FAULT, EI_HOSTUNREACH, EI_INTR, EI_MSGSIZE, EI_NETDOWN, EI_NETUNREACH, EI_NOBUFS, EI_NOTSOCK, EI_OPNOTSUPP, EI_PIPE};


int geterrorno();

#endif

