/*
 * Copyright (C) 2008 The Trustees of Columbia University in the City of New York. 
 * All rights reserved.
 *
 * See the file LICENSE included in this distribution for details.
 */


#include "threadi.h"

#ifdef _LINUX_
#include <errno.h>
#include <pthread.h>
#endif

#ifdef WIN32
#include <winsock.h>
#endif


int32 CreateThreadI(void *(*start_routine)(void*), void *arg) {

#ifdef _LINUX_
  pthread_t tid;
  pthread_create(&tid, NULL, start_routine, arg);
  return (int32)tid;
#endif

#ifdef WIN32
  DWORD tid;
  HANDLE t_h;
  t_h = CreateThread( 0, 0, (LPTHREAD_START_ROUTINE)start_routine, arg, 0, &tid);
  return (int32)t_h;
#endif

  return 1;
}//CreateThreadI



int32 WaitForThread(int32 iThread) {

#ifdef _LINUX_
  pthread_join((pthread_t)iThread, NULL);
#endif

#ifdef WIN32  
  WaitForMultipleObjects( 1, (HANDLE*)&iThread, TRUE, INFINITE);
#endif

  return 1;
}//WaitForThread



int32 ExitThread(int32 iThread) {

#ifdef _LINUX_
  pthread_exit(NULL);
#endif

#ifdef WIN32

#endif

  return 1;
}//ExitThread


int32 DetachThread(int32 iThread) {

#ifdef _LINUX_
  return pthread_detach(iThread);
#endif

  return 0;
}//DetachThread


int32 KillThread(int32 iThread) {

#ifdef _LINUX_
  //pthread_kill((pthread_t)iThread,0);
#endif
  
  return 1;
}//KillThread


int32 SuspendThread(int32 iThread) {

#ifdef _LINUX_
  //  pthread_suspend((pthread_t)iThread);
#endif
  
  return 1;
}//SuspendThread


int32 ResumeThread(int32 iThread) {
 
#ifdef _LINUX_
  //  pthread_resume((pthread_t)iThread);
#endif
  
  return 1;
}//ResumeThread



////////////////////////////////////////
// Mutex functions
////////////////////////////////////////

/*
  1 successful
  -1 failure
*/
int32 allocmutex_i(struct mutex_i* pstMutex) {  

#ifdef WIN32
  pstMutex->mutexI=CreateMutex(NULL, FALSE, NULL);
  if(NULL != pstMutex->mutexI) {
    return 1;
  }
#endif

#ifdef _LINUX_
  int32 ret;
  ret=pthread_mutex_init(&pstMutex->mutexI, NULL);
  //if non-zero, there was error
  if(0 == ret) {
    return 1;
  }
#endif

  //we could not allocate the mutex
  return -1;
}//allocmutex_i


/* 1 acquire lock
  -1 cannot acquire lock
 */
int32 lockmutex_i(struct mutex_i* pstMutex) {

  int32 ret=0;

#ifdef WIN32
  ret=WaitForSingleObject( pstMutex->mutexI, INFINITE );
  if(ret==WAIT_OBJECT_0) {
    return 1;
  }
#endif

#ifdef _LINUX_
  ret=pthread_mutex_lock(&pstMutex->mutexI);
  if(0==ret) {
    return 1;
  }
#endif

  return -1;
}//lockmutex_i


/*
  1 successful
  0 cannot acquire lock
  -1 failure
 */
int32 trylockmutex_i(struct mutex_i* pstMutex) {
  int32 ret=0;

#ifdef WIN32
  ret=WaitForSingleObject(pstMutex->mutexI, 0 );
  switch(ret) {
  case WAIT_FAILED:   //there was an error
    return -1;
  case WAIT_OBJECT_0: //mutex acquired
    return 1;
  default:            //mutex cannot be acquired
    return 0;
  }//switch
#endif


#ifdef _LINUX_
  ret=pthread_mutex_trylock(&pstMutex->mutexI);
  if(0==ret) {
    return 1;
  } else if(EBUSY==ret) {
    return 0;
  }
#endif
  
  return -1;
}//trylockmutex_i


/*
  1 successful
  -1 error
 */
int32 unlockmutex_i(struct mutex_i* pstMutex) {
  int32 ret=0;

#ifdef WIN32
  ret=ReleaseMutex( pstMutex->mutexI);
  //non-zero is success, 0 is failure
  if(0!=ret) {
    return 1;
  }
#endif

#ifdef _LINUX_
  //0 is success, 1 is failure
  ret=pthread_mutex_unlock(&pstMutex->mutexI);
  if(0==ret) {
    return 1;
  }
#endif

  return -1;
}//unlockmutex_i


/*
  1 successful
  -1 failure
 */
int32 deallocmutex_i(struct mutex_i* pstMutex) {
  int32 ret=0;

#ifdef WIN32
  ret=CloseHandle(pstMutex->mutexI);
  //non-zero is success, 0 is failure for pthread functions
  if(0 != ret) {
    return 1;
  }
#endif

#ifdef _LINUX_
  ret=pthread_mutex_destroy(&pstMutex->mutexI);
  //0 is success rest is failure for pthread functions
  if(0==ret) {
    return 1;
  }
#endif

  return 0;
}//deallocmutex_i

